# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import math
import os
import re
import tempfile
from dataclasses import dataclass


from .... import ModelFormat, UltraInferModel
from .... import c_lib_wrap as C
from ....py_only import PyOnlyProcessorChain
from ....py_only.vision import PyOnlyVisionModel
from ....py_only.vision import processors as P
from ....utils.misc import load_config
from ...common import ProcessorManager
from .utils.ser_vi_layoutxlm.transforms import *
from .utils.ser_vi_layoutxlm.vqa_utils import *
from .utils.ser_vi_layoutxlm.operators import *


def sort_boxes(boxes):
    return C.vision.ocr.sort_boxes(boxes)


class UVDocPreprocessor(ProcessorManager):
    def __init__(self):
        """Create a preprocessor for UVDoc Model"""
        super(UVDocPreprocessor, self).__init__()
        self._manager = C.vision.ocr.UVDocPreprocessor()

    def set_normalize(self, mean, std, is_scale):
        """Set preprocess normalize parameters, please call this API to
           customize the normalize parameters, otherwise it will use the default
           normalize parameters.
        :param: mean: (list of float) mean values
        :param: std: (list of float) std values
        :param: is_scale: (boolean) whether to scale
        """
        self._manager.set_normalize(mean, std, is_scale)

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()


class UVDocPostprocessor:
    def __init__(self):
        """Create a postprocessor for UVDoc Model"""
        self._postprocessor = C.vision.ocr.UVDocPostprocessor()

    def run(self, runtime_results):
        """Postprocess the runtime results for UVDoc Model
        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)


class UVDocWarpper(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load OCR recognition model provided by PaddleOCR

        :param model_file: (str)Path of model file, e.g ./ch_PP-OCRv3_rec_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_PP-OCRv3_rec_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(UVDocWarpper, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.UVDocWarpper()
            self._runnable = False
        else:
            self._model = C.vision.ocr.UVDocWarpper(
                model_file, params_file, self._runtime_option, model_format
            )
            assert self.initialized, "UVDocWarpper initialize failed."
            self._runnable = True

    def clone(self):
        """Clone OCR recognition model object
        :return: a new OCR recognition model object
        """

        class UVDocWarpperClone(UVDocWarpper):
            def __init__(self, model):
                self._model = model

        clone_model = UVDocWarpperClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: rec_text, rec_score
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of rec_text, list of rec_score
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @preprocessor.setter
    def preprocessor(self, value):
        self._model.preprocessor = value

    @property
    def postprocessor(self):
        return self._model.postprocessor

    @postprocessor.setter
    def postprocessor(self, value):
        self._model.postprocessor = value


class DBDetectorPreprocessor(ProcessorManager):
    def __init__(self):
        """
        Create a preprocessor for DBDetectorModel
        """
        super(DBDetectorPreprocessor, self).__init__()
        self._manager = C.vision.ocr.DBDetectorPreprocessor()

    @property
    def max_side_len(self):
        """Get max_side_len value."""
        return self._manager.max_side_len

    @max_side_len.setter
    def max_side_len(self, value):
        """Set max_side_len value.
        :param: value: (int) max_side_len value
        """
        assert isinstance(
            value, int
        ), "The value to set `max_side_len` must be type of int."
        self._manager.max_side_len = value

    def set_normalize(self, mean, std, is_scale):
        """Set preprocess normalize parameters, please call this API to
           customize the normalize parameters, otherwise it will use the default
           normalize parameters.
        :param: mean: (list of float) mean values
        :param: std: (list of float) std values
        :param: is_scale: (boolean) whether to scale
        """
        self._manager.set_normalize(mean, std, is_scale)

    @property
    def static_shape_infer(self):
        return self._manager.static_shape_infer

    @static_shape_infer.setter
    def static_shape_infer(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `static_shape_infer` must be type of bool."
        self._manager.static_shape_infer = value

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()


class DBDetectorPostprocessor:
    def __init__(self):
        """
        Create a postprocessor for DBDetectorModel
        """
        self._postprocessor = C.vision.ocr.DBDetectorPostprocessor()

    def run(self, runtime_results, batch_det_img_info):
        """Postprocess the runtime results for DBDetectorModel

        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :param: batch_det_img_info: (list of std::array<int, 4>)The output of det_preprocessor
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results, batch_det_img_info)

    @property
    def det_db_thresh(self):
        """
        Return the det_db_thresh of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_thresh

    @det_db_thresh.setter
    def det_db_thresh(self, value):
        """Set the det_db_thresh for DBDetectorPostprocessor

        :param: value : the det_db_thresh value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_thresh` must be type of float."
        self._postprocessor.det_db_thresh = value

    @property
    def det_db_box_thresh(self):
        """
        Return the det_db_box_thresh of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_box_thresh

    @det_db_box_thresh.setter
    def det_db_box_thresh(self, value):
        """Set the det_db_box_thresh for DBDetectorPostprocessor

        :param: value : the det_db_box_thresh value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_box_thresh` must be type of float."
        self._postprocessor.det_db_box_thresh = value

    @property
    def det_db_unclip_ratio(self):
        """
        Return the det_db_unclip_ratio of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_unclip_ratio

    @det_db_unclip_ratio.setter
    def det_db_unclip_ratio(self, value):
        """Set the det_db_unclip_ratio for DBDetectorPostprocessor

        :param: value : the det_db_unclip_ratio value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_unclip_ratio` must be type of float."
        self._postprocessor.det_db_unclip_ratio = value

    @property
    def det_db_score_mode(self):
        """
        Return the det_db_score_mode of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_score_mode

    @property
    def det_db_box_type(self):
        """
        Return the det_db_score_mode of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_box_type

    @det_db_box_type.setter
    def det_db_box_type(self, value):
        """Set the det_db_score_mode for DBDetectorPostprocessor

        :param: value : the det_db_score_mode value
        """
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._postprocessor.det_db_box_type = value

    @det_db_score_mode.setter
    def det_db_score_mode(self, value):
        """Set the det_db_score_mode for DBDetectorPostprocessor

        :param: value : the det_db_score_mode value
        """
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._postprocessor.det_db_score_mode = value

    @property
    def use_dilation(self):
        """
        Return the use_dilation of DBDetectorPostprocessor
        """
        return self._postprocessor.use_dilation

    @use_dilation.setter
    def use_dilation(self, value):
        """Set the use_dilation for DBDetectorPostprocessor

        :param: value : the use_dilation value
        """
        assert isinstance(
            value, bool
        ), "The value to set `use_dilation` must be type of bool."
        self._postprocessor.use_dilation = value


class DBCURVEDetectorPostprocessor:
    def __init__(self):
        """
        Create a postprocessor for DBDetectorModel
        """
        self._postprocessor = C.vision.ocr.DBCURVEDetectorPostprocessor()

    def run(self, runtime_results, batch_det_img_info):
        """Postprocess the runtime results for DBDetectorModel

        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :param: batch_det_img_info: (list of std::array<int, 4>)The output of det_preprocessor
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results, batch_det_img_info)

    @property
    def det_db_thresh(self):
        """
        Return the det_db_thresh of DBCURVEDetectorPostprocessor
        """
        return self._postprocessor.det_db_thresh

    @det_db_thresh.setter
    def det_db_thresh(self, value):
        """Set the det_db_thresh for DBCURVEDetectorPostprocessor

        :param: value : the det_db_thresh value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_thresh` must be type of float."
        self._postprocessor.det_db_thresh = value

    @property
    def det_db_box_thresh(self):
        """
        Return the det_db_box_thresh of DBCURVEDetectorPostprocessor
        """
        return self._postprocessor.det_db_box_thresh

    @det_db_box_thresh.setter
    def det_db_box_thresh(self, value):
        """Set the det_db_box_thresh for DBCURVEDetectorPostprocessor

        :param: value : the det_db_box_thresh value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_box_thresh` must be type of float."
        self._postprocessor.det_db_box_thresh = value

    @property
    def det_db_unclip_ratio(self):
        """
        Return the det_db_unclip_ratio of DBCURVEDetectorPostprocessor
        """
        return self._postprocessor.det_db_unclip_ratio

    @det_db_unclip_ratio.setter
    def det_db_unclip_ratio(self, value):
        """Set the det_db_unclip_ratio for DBCURVEDetectorPostprocessor

        :param: value : the det_db_unclip_ratio value
        """
        assert isinstance(
            value, float
        ), "The value to set `det_db_unclip_ratio` must be type of float."
        self._postprocessor.det_db_unclip_ratio = value

    @property
    def det_db_score_mode(self):
        """
        Return the det_db_score_mode of DBCURVEDetectorPostprocessor
        """
        return self._postprocessor.det_db_score_mode

    @property
    def det_db_box_type(self):
        """
        Return the det_db_score_mode of DBDetectorPostprocessor
        """
        return self._postprocessor.det_db_box_type

    @det_db_box_type.setter
    def det_db_box_type(self, value):
        """Set the det_db_score_mode for DBDetectorPostprocessor

        :param: value : the det_db_score_mode value
        """
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._postprocessor.det_db_box_type = value

    @det_db_score_mode.setter
    def det_db_score_mode(self, value):
        """Set the det_db_score_mode for DBDetectorPostprocessor

        :param: value : the det_db_score_mode value
        """
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._postprocessor.det_db_score_mode = value

    @property
    def use_dilation(self):
        """
        Return the use_dilation of DBDetectorPostprocessor
        """
        return self._postprocessor.use_dilation

    @use_dilation.setter
    def use_dilation(self, value):
        """Set the use_dilation for DBDetectorPostprocessor

        :param: value : the use_dilation value
        """
        assert isinstance(
            value, bool
        ), "The value to set `use_dilation` must be type of bool."
        self._postprocessor.use_dilation = value


class DBDetector(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load OCR detection model provided by PaddleOCR.

        :param model_file: (str)Path of model file, e.g ./ch_PP-OCRv3_det_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_PP-OCRv3_det_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(DBDetector, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.DBDetector()
            self._runnable = False
        else:
            self._model = C.vision.ocr.DBDetector(
                model_file, params_file, self._runtime_option, model_format
            )
            assert self.initialized, "DBDetector initialize failed."
            self._runnable = True

    def clone(self):
        """Clone OCR detection model object

        :return: a new OCR detection model object
        """

        class DBDetectorClone(DBDetector):
            def __init__(self, model):
                self._model = model

        clone_model = DBDetectorClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: boxes
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: batch_boxes
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @property
    def postprocessor(self):
        return self._model.postprocessor

    # Det Preprocessor Property
    @property
    def max_side_len(self):
        return self._model.preprocessor.max_side_len

    @max_side_len.setter
    def max_side_len(self, value):
        assert isinstance(
            value, int
        ), "The value to set `max_side_len` must be type of int."
        self._model.preprocessor.max_side_len = value

    # Det Ppstprocessor Property
    @property
    def det_db_thresh(self):
        return self._model.postprocessor.det_db_thresh

    @det_db_thresh.setter
    def det_db_thresh(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_thresh` must be type of float."
        self._model.postprocessor.det_db_thresh = value

    @property
    def det_db_box_thresh(self):
        return self._model.postprocessor.det_db_box_thresh

    @det_db_box_thresh.setter
    def det_db_box_thresh(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_box_thresh` must be type of float."
        self._model.postprocessor.det_db_box_thresh = value

    @property
    def det_db_unclip_ratio(self):
        return self._model.postprocessor.det_db_unclip_ratio

    @det_db_unclip_ratio.setter
    def det_db_unclip_ratio(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_unclip_ratio` must be type of float."
        self._model.postprocessor.det_db_unclip_ratio = value

    @property
    def det_db_box_type(self):
        return self._model.postprocessor.det_db_box_type

    @det_db_box_type.setter
    def det_db_box_type(self, value):
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._model.postprocessor.det_db_box_type = value

    @property
    def det_db_score_mode(self):
        return self._model.postprocessor.det_db_score_mode

    @det_db_score_mode.setter
    def det_db_score_mode(self, value):
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._model.postprocessor.det_db_score_mode = value

    @property
    def use_dilation(self):
        return self._model.postprocessor.use_dilation

    @use_dilation.setter
    def use_dilation(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `use_dilation` must be type of bool."
        self._model.postprocessor.use_dilation = value


class DBCURVEDetector(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load OCR detection model provided by PaddleOCR.

        :param model_file: (str)Path of model file, e.g ./ch_PP-OCRv3_det_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_PP-OCRv3_det_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(DBCURVEDetector, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.DBCURVEDetector()
            self._runnable = False
        else:
            self._model = C.vision.ocr.DBCURVEDetector(
                model_file, params_file, self._runtime_option, model_format
            )
            assert self.initialized, "DBCURVEDetector initialize failed."
            self._runnable = True

    def clone(self):
        """Clone OCR detection model object

        :return: a new OCR detection model object
        """

        class DBCURVEDetectorClone(DBCURVEDetector):
            def __init__(self, model):
                self._model = model

        clone_model = DBCURVEDetectorClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: boxes
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: batch_boxes
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @property
    def postprocessor(self):
        return self._model.postprocessor

    # Det Preprocessor Property
    @property
    def max_side_len(self):
        return self._model.preprocessor.max_side_len

    @max_side_len.setter
    def max_side_len(self, value):
        assert isinstance(
            value, int
        ), "The value to set `max_side_len` must be type of int."
        self._model.preprocessor.max_side_len = value

    # Det Ppstprocessor Property
    @property
    def det_db_thresh(self):
        return self._model.postprocessor.det_db_thresh

    @det_db_thresh.setter
    def det_db_thresh(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_thresh` must be type of float."
        self._model.postprocessor.det_db_thresh = value

    @property
    def det_db_box_thresh(self):
        return self._model.postprocessor.det_db_box_thresh

    @det_db_box_thresh.setter
    def det_db_box_thresh(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_box_thresh` must be type of float."
        self._model.postprocessor.det_db_box_thresh = value

    @property
    def det_db_unclip_ratio(self):
        return self._model.postprocessor.det_db_unclip_ratio

    @det_db_unclip_ratio.setter
    def det_db_unclip_ratio(self, value):
        assert isinstance(
            value, float
        ), "The value to set `det_db_unclip_ratio` must be type of float."
        self._model.postprocessor.det_db_unclip_ratio = value

    @property
    def det_db_box_type(self):
        return self._model.postprocessor.det_db_box_type

    @det_db_box_type.setter
    def det_db_box_type(self, value):
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._model.postprocessor.det_db_box_type = value

    @property
    def det_db_score_mode(self):
        return self._model.postprocessor.det_db_score_mode

    @det_db_score_mode.setter
    def det_db_score_mode(self, value):
        assert isinstance(
            value, str
        ), "The value to set `det_db_score_mode` must be type of str."
        self._model.postprocessor.det_db_score_mode = value

    @property
    def use_dilation(self):
        return self._model.postprocessor.use_dilation

    @use_dilation.setter
    def use_dilation(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `use_dilation` must be type of bool."
        self._model.postprocessor.use_dilation = value


class ClassifierPreprocessor(ProcessorManager):
    def __init__(self):
        """Create a preprocessor for ClassifierModel"""
        super(ClassifierPreprocessor, self).__init__()
        self._manager = C.vision.ocr.ClassifierPreprocessor()

    def set_normalize(self, mean, std, is_scale):
        """Set preprocess normalize parameters, please call this API to
           customize the normalize parameters, otherwise it will use the default
           normalize parameters.
        :param: mean: (list of float) mean values
        :param: std: (list of float) std values
        :param: is_scale: (boolean) whether to scale
        """
        self._manager.set_normalize(mean, std, is_scale)

    @property
    def cls_image_shape(self):
        return self._manager.cls_image_shape

    @cls_image_shape.setter
    def cls_image_shape(self, value):
        assert isinstance(
            value, list
        ), "The value to set `cls_image_shape` must be type of list."
        self._manager.cls_image_shape = value

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()


class ClassifierPostprocessor:
    def __init__(self):
        """Create a postprocessor for ClassifierModel"""
        self._postprocessor = C.vision.ocr.ClassifierPostprocessor()

    def run(self, runtime_results):
        """Postprocess the runtime results for ClassifierModel
        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)

    @property
    def cls_thresh(self):
        """
        Return the cls_thresh of ClassifierPostprocessor
        """
        return self._postprocessor.cls_thresh

    @cls_thresh.setter
    def cls_thresh(self, value):
        """Set the cls_thresh for ClassifierPostprocessor

        :param: value: the value of cls_thresh
        """
        assert isinstance(
            value, float
        ), "The value to set `cls_thresh` must be type of float."
        self._postprocessor.cls_thresh = value


class Classifier(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load OCR classification model provided by PaddleOCR.

        :param model_file: (str)Path of model file, e.g ./ch_ppocr_mobile_v2.0_cls_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_ppocr_mobile_v2.0_cls_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(Classifier, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.Classifier()
            self._runnable = False
        else:
            self._model = C.vision.ocr.Classifier(
                model_file, params_file, self._runtime_option, model_format
            )
            assert self.initialized, "Classifier initialize failed."
            self._runnable = True

    def clone(self):
        """Clone OCR classification model object
        :return: a new OCR classification model object
        """

        class ClassifierClone(Classifier):
            def __init__(self, model):
                self._model = model

        clone_model = ClassifierClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: cls_label, cls_score
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of cls_label, list of cls_score
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @preprocessor.setter
    def preprocessor(self, value):
        self._model.preprocessor = value

    @property
    def postprocessor(self):
        return self._model.postprocessor

    @postprocessor.setter
    def postprocessor(self, value):
        self._model.postprocessor = value

    @property
    def cls_image_shape(self):
        return self._model.preprocessor.cls_image_shape

    @cls_image_shape.setter
    def cls_image_shape(self, value):
        assert isinstance(
            value, list
        ), "The value to set `cls_image_shape` must be type of list."
        self._model.preprocessor.cls_image_shape = value

    # Cls Postprocessor Property
    @property
    def cls_thresh(self):
        return self._model.postprocessor.cls_thresh

    @cls_thresh.setter
    def cls_thresh(self, value):
        assert isinstance(
            value, float
        ), "The value to set `cls_thresh` must be type of float."
        self._model.postprocessor.cls_thresh = value


class RecognizerPreprocessor(ProcessorManager):
    def __init__(self):
        """Create a preprocessor for RecognizerModel"""
        super(RecognizerPreprocessor, self).__init__()
        self._manager = C.vision.ocr.RecognizerPreprocessor()

    @property
    def static_shape_infer(self):
        return self._manager.static_shape_infer

    @static_shape_infer.setter
    def static_shape_infer(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `static_shape_infer` must be type of bool."
        self._manager.static_shape_infer = value

    def set_normalize(self, mean, std, is_scale):
        """Set preprocess normalize parameters, please call this API to
           customize the normalize parameters, otherwise it will use the default
           normalize parameters.
        :param: mean: (list of float) mean values
        :param: std: (list of float) std values
        :param: is_scale: (boolean) whether to scale
        """
        self._manager.set_normalize(mean, std, is_scale)

    @property
    def rec_image_shape(self):
        return self._manager.rec_image_shape

    @rec_image_shape.setter
    def rec_image_shape(self, value):
        assert isinstance(
            value, list
        ), "The value to set `rec_image_shape` must be type of list."
        self._manager.rec_image_shape = value

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()


class RecognizerPostprocessor:
    def __init__(self, label_path):
        """Create a postprocessor for RecognizerModel
        :param label_path: (str)Path of label file
        """
        self._postprocessor = C.vision.ocr.RecognizerPostprocessor(label_path)

    def run(self, runtime_results):
        """Postprocess the runtime results for RecognizerModel
        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)


class Recognizer(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        label_path="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load OCR recognition model provided by PaddleOCR

        :param model_file: (str)Path of model file, e.g ./ch_PP-OCRv3_rec_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_PP-OCRv3_rec_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param label_path: (str)Path of label file used by OCR recognition model. e.g ./ppocr_keys_v1.txt
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(Recognizer, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.Recognizer()
            self._runnable = False
        else:
            self._model = C.vision.ocr.Recognizer(
                model_file, params_file, label_path, self._runtime_option, model_format
            )
            assert self.initialized, "Recognizer initialize failed."
            self._runnable = True

    def clone(self):
        """Clone OCR recognition model object
        :return: a new OCR recognition model object
        """

        class RecognizerClone(Recognizer):
            def __init__(self, model):
                self._model = model

        clone_model = RecognizerClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: rec_text, rec_score
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of rec_text, list of rec_score
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @preprocessor.setter
    def preprocessor(self, value):
        self._model.preprocessor = value

    @property
    def postprocessor(self):
        return self._model.postprocessor

    @postprocessor.setter
    def postprocessor(self, value):
        self._model.postprocessor = value

    @property
    def static_shape_infer(self):
        return self._model.preprocessor.static_shape_infer

    @static_shape_infer.setter
    def static_shape_infer(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `static_shape_infer` must be type of bool."
        self._model.preprocessor.static_shape_infer = value

    @property
    def rec_image_shape(self):
        return self._model.preprocessor.rec_image_shape

    @rec_image_shape.setter
    def rec_image_shape(self, value):
        assert isinstance(
            value, list
        ), "The value to set `rec_image_shape` must be type of list."
        self._model.preprocessor.rec_image_shape = value


class StructureV2TablePreprocessor:
    def __init__(self):
        """Create a preprocessor for StructureV2Table Model"""
        self._preprocessor = C.vision.ocr.StructureV2TablePreprocessor()

    def run(self, input_ims):
        """Preprocess input images for StructureV2TableModel
        :param: input_ims: (list of numpy.ndarray)The input image
        :return: list of FDTensor
        """
        return self._preprocessor.run(input_ims)


class StructureV2TablePostprocessor:
    def __init__(self, dict_path):
        """Create a postprocessor for StructureV2Table Model"""
        self._postprocessor = C.vision.ocr.StructureV2TablePostprocessor(dict_path)

    def run(self, runtime_results):
        """Postprocess the runtime results for StructureV2Table Model
        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)


class StructureV2Table(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        table_char_dict_path="",
        box_shape="ori",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load StructureV2Table model provided by PP-StructureV2.

        :param model_file: (str)Path of model file, e.g ./ch_ppocr_mobile_v2.0_cls_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ch_ppocr_mobile_v2.0_cls_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param table_char_dict_path: (str)Path of table_char_dict file, e.g ../ppocr/utils/dict/table_structure_dict_ch.txt
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(StructureV2Table, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.StructureV2Table()
            self._runnable = False
        else:
            self._model = C.vision.ocr.StructureV2Table(
                model_file,
                params_file,
                table_char_dict_path,
                box_shape,
                self._runtime_option,
                model_format,
            )
            assert self.initialized, "Classifier initialize failed."
            self._runnable = True

    def clone(self):
        """Clone StructureV2Table model object
        :return: a new StructureV2Table model object
        """

        class StructureV2TableClone(StructureV2Table):
            def __init__(self, model):
                self._model = model

        clone_model = StructureV2TableClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: bbox, structure
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of bbox list, list of structure
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @preprocessor.setter
    def preprocessor(self, value):
        self._model.preprocessor = value

    @property
    def postprocessor(self):
        return self._model.postprocessor

    @postprocessor.setter
    def postprocessor(self, value):
        self._model.postprocessor = value


class StructureV2LayoutPreprocessor:
    def __init__(self):
        """Create a preprocessor for StructureV2Layout Model"""
        self._preprocessor = C.vision.ocr.StructureV2LayoutPreprocessor()

    def run(self, input_ims):
        """Preprocess input images for StructureV2Layout Model
        :param: input_ims: (list of numpy.ndarray)The input image
        :return: list of FDTensor
        """
        return self._preprocessor.run(input_ims)


class StructureV2LayoutPostprocessor:
    def __init__(self):
        """Create a postprocessor for StructureV2Layout Model"""
        self._postprocessor = C.vision.ocr.StructureV2LayoutPostprocessor()

    def run(self, runtime_results):
        """Postprocess the runtime results for StructureV2Layout Model
        :param: runtime_results: (list of FDTensor or list of pyArray)The output FDTensor results from runtime
        :return: list of Result(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)


class StructureV2Layout(UltraInferModel):
    def __init__(
        self,
        model_file="",
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load StructureV2Layout model provided by PP-StructureV2.

        :param model_file: (str)Path of model file, e.g ./picodet_lcnet_x1_0_fgd_layout_infer/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./picodet_lcnet_x1_0_fgd_layout_infer/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(StructureV2Layout, self).__init__(runtime_option)

        if len(model_file) == 0:
            self._model = C.vision.ocr.StructureV2Layout()
            self._runnable = False
        else:
            self._model = C.vision.ocr.StructureV2Layout(
                model_file, params_file, self._runtime_option, model_format
            )
            assert self.initialized, "StructureV2Layout model initialize failed."
            self._runnable = True

    def clone(self):
        """Clone StructureV2Layout model object
        :return: a new StructureV2Table model object
        """

        class StructureV2LayoutClone(StructureV2Layout):
            def __init__(self, model):
                self._model = model

        clone_model = StructureV2LayoutClone(self._model.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: bboxes
        """
        if self._runnable:
            return self._model.predict(input_image)
        return False

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of bboxes list
        """
        if self._runnable:
            return self._model.batch_predict(images)
        return False

    @property
    def preprocessor(self):
        return self._model.preprocessor

    @preprocessor.setter
    def preprocessor(self, value):
        self._model.preprocessor = value

    @property
    def postprocessor(self):
        return self._model.postprocessor

    @postprocessor.setter
    def postprocessor(self, value):
        self._model.postprocessor = value


class PPOCRv4(UltraInferModel):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        """Construct a pipeline with text detector, direction classifier and text recognizer models

        :param det_model: (UltraInferModel) The detection model object created by ultra_infer.vision.ocr.DBDetector.
        :param cls_model: (UltraInferModel) The classification model object created by ultra_infer.vision.ocr.Classifier.
        :param rec_model: (UltraInferModel) The recognition model object created by ultra_infer.vision.ocr.Recognizer.
        """
        assert (
            det_model is not None and rec_model is not None
        ), "The det_model and rec_model cannot be None."

        self.det_model = det_model
        self.rec_model = rec_model
        self.cls_model = cls_model

        if cls_model is None:
            self.system_ = C.vision.ocr.PPOCRv4(det_model._model, rec_model._model)
        else:
            self.system_ = C.vision.ocr.PPOCRv4(
                det_model._model, cls_model._model, rec_model._model
            )

    def clone(self):
        """Clone PPOCRv4 pipeline object
        :return: a new PPOCRv4 pipeline object
        """

        class PPOCRv4Clone(PPOCRv4):
            def __init__(self, system):
                self.system_ = system

        clone_model = PPOCRv4Clone(self.system_.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: OCRResult
        """
        return self.system_.predict(input_image)

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: OCRBatchResult
        """
        return self.system_.batch_predict(images)

    @property
    def cls_batch_size(self):
        return self.system_.cls_batch_size

    @cls_batch_size.setter
    def cls_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `cls_batch_size` must be type of int."
        self.system_.cls_batch_size = value

    @property
    def rec_batch_size(self):
        return self.system_.rec_batch_size

    @rec_batch_size.setter
    def rec_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `rec_batch_size` must be type of int."
        self.system_.rec_batch_size = value


class PPOCRSystemv4(PPOCRv4):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        logging.warning(
            "DEPRECATED: fd.vision.ocr.PPOCRSystemv4 is deprecated, "
            "please use fd.vision.ocr.PPOCRv4 instead."
        )
        super(PPOCRSystemv4, self).__init__(det_model, cls_model, rec_model)

    def predict(self, input_image):
        return super(PPOCRSystemv4, self).predict(input_image)


class PPOCRv3(UltraInferModel):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        """Construct a pipeline with text detector, direction classifier and text recognizer models

        :param det_model: (UltraInferModel) The detection model object created by ultra_infer.vision.ocr.DBDetector.
        :param cls_model: (UltraInferModel) The classification model object created by ultra_infer.vision.ocr.Classifier.
        :param rec_model: (UltraInferModel) The recognition model object created by ultra_infer.vision.ocr.Recognizer.
        """
        assert (
            det_model is not None and rec_model is not None
        ), "The det_model and rec_model cannot be None."
        if cls_model is None:
            self.system_ = C.vision.ocr.PPOCRv3(det_model._model, rec_model._model)
        else:
            self.system_ = C.vision.ocr.PPOCRv3(
                det_model._model, cls_model._model, rec_model._model
            )

    def clone(self):
        """Clone PPOCRv3 pipeline object
        :return: a new PPOCRv3 pipeline object
        """

        class PPOCRv3Clone(PPOCRv3):
            def __init__(self, system):
                self.system_ = system

        clone_model = PPOCRv3Clone(self.system_.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: OCRResult
        """
        return self.system_.predict(input_image)

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: OCRBatchResult
        """
        return self.system_.batch_predict(images)

    @property
    def cls_batch_size(self):
        return self.system_.cls_batch_size

    @cls_batch_size.setter
    def cls_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `cls_batch_size` must be type of int."
        self.system_.cls_batch_size = value

    @property
    def rec_batch_size(self):
        return self.system_.rec_batch_size

    @rec_batch_size.setter
    def rec_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `rec_batch_size` must be type of int."
        self.system_.rec_batch_size = value


class PPOCRSystemv3(PPOCRv3):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        logging.warning(
            "DEPRECATED: fd.vision.ocr.PPOCRSystemv3 is deprecated, "
            "please use fd.vision.ocr.PPOCRv3 instead."
        )
        super(PPOCRSystemv3, self).__init__(det_model, cls_model, rec_model)

    def predict(self, input_image):
        return super(PPOCRSystemv3, self).predict(input_image)


class PPOCRv2(UltraInferModel):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        """Construct a pipeline with text detector, direction classifier and text recognizer models

        :param det_model: (UltraInferModel) The detection model object created by ultra_infer.vision.ocr.DBDetector.
        :param cls_model: (UltraInferModel) The classification model object created by ultra_infer.vision.ocr.Classifier.
        :param rec_model: (UltraInferModel) The recognition model object created by ultra_infer.vision.ocr.Recognizer.
        """
        assert (
            det_model is not None and rec_model is not None
        ), "The det_model and rec_model cannot be None."
        if cls_model is None:
            self.system_ = C.vision.ocr.PPOCRv2(det_model._model, rec_model._model)
        else:
            self.system_ = C.vision.ocr.PPOCRv2(
                det_model._model, cls_model._model, rec_model._model
            )

    def clone(self):
        """Clone PPOCRv3 pipeline object
        :return: a new PPOCRv3 pipeline object
        """

        class PPOCRv2Clone(PPOCRv2):
            def __init__(self, system):
                self.system_ = system

        clone_model = PPOCRv2Clone(self.system_.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image

        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: OCRResult
        """
        return self.system_.predict(input_image)

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: OCRBatchResult
        """

        return self.system_.batch_predict(images)

    @property
    def cls_batch_size(self):
        return self.system_.cls_batch_size

    @cls_batch_size.setter
    def cls_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `cls_batch_size` must be type of int."
        self.system_.cls_batch_size = value

    @property
    def rec_batch_size(self):
        return self.system_.rec_batch_size

    @rec_batch_size.setter
    def rec_batch_size(self, value):
        assert isinstance(
            value, int
        ), "The value to set `rec_batch_size` must be type of int."
        self.system_.rec_batch_size = value


class PPOCRSystemv2(PPOCRv2):
    def __init__(self, det_model=None, cls_model=None, rec_model=None):
        logging.warning(
            "DEPRECATED: fd.vision.ocr.PPOCRSystemv2 is deprecated, "
            "please use fd.vision.ocr.PPOCRv2 instead."
        )
        super(PPOCRSystemv2, self).__init__(det_model, cls_model, rec_model)

    def predict(self, input_image):
        return super(PPOCRSystemv2, self).predict(input_image)


class PPStructureV2Table(UltraInferModel):
    def __init__(self, det_model=None, rec_model=None, table_model=None):
        """Construct a pipeline with text detector, text recognizer and table recognizer models

        :param det_model: (UltraInferModel) The detection model object created by ultra_infer.vision.ocr.DBDetector.
        :param rec_model: (UltraInferModel) The recognition model object created by ultra_infer.vision.ocr.Recognizer.
        :param table_model: (UltraInferModel) The table recognition model object created by ultra_infer.vision.ocr.Table.
        """
        assert (
            det_model is not None and rec_model is not None and table_model is not None
        ), "The det_model, rec_model and table_model cannot be None."
        self.system_ = C.vision.ocr.PPStructureV2Table(
            det_model._model,
            rec_model._model,
            table_model._model,
        )

    def clone(self):
        """Clone PPStructureV2Table pipeline object
        :return: a new PPStructureV2Table pipeline object
        """

        class PPStructureV2TableClone(PPStructureV2Table):
            def __init__(self, system):
                self.system_ = system

        clone_model = PPStructureV2TableClone(self.system_.clone())
        return clone_model

    def predict(self, input_image):
        """Predict an input image

        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: OCRResult
        """
        return self.system_.predict(input_image)

    def batch_predict(self, images):
        """Predict a batch of input image
        :param images: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: OCRBatchResult
        """

        return self.system_.batch_predict(images)


class PPStructureV2TableSystem(PPStructureV2Table):
    def __init__(self, det_model=None, rec_model=None, table_model=None):
        logging.warning(
            "DEPRECATED: fd.vision.ocr.PPStructureV2TableSystem is deprecated, "
            "please use fd.vision.ocr.PPStructureV2Table instead."
        )
        super(PPStructureV2TableSystem, self).__init__(
            det_model, rec_model, table_model
        )

    def predict(self, input_image):
        return super(PPStructureV2TableSystem, self).predict(input_image)


class StructureV2SERViLayoutXLMModelPreprocessor:
    def __init__(self, ser_dict_path, use_gpu=True):
        """Create a preprocessor for Ser-Vi-LayoutXLM model.
        :param: ser_dict_path: (str) class file path
        :param: use_gpu: (bool) whether use gpu to OCR process
        """
        self._manager = None
        from paddleocr import PaddleOCR

        self.ocr_engine = PaddleOCR(
            use_angle_cls=False,
            det_model_dir=None,
            rec_model_dir=None,
            show_log=False,
            use_gpu=use_gpu,
        )

        pre_process_list = [
            {
                "VQATokenLabelEncode": {
                    "class_path": ser_dict_path,
                    "contains_re": False,
                    "ocr_engine": self.ocr_engine,
                    "order_method": "tb-yx",
                }
            },
            {"VQATokenPad": {"max_seq_len": 512, "return_attention_mask": True}},
            {"VQASerTokenChunk": {"max_seq_len": 512, "return_attention_mask": True}},
            {"Resize": {"size": [224, 224]}},
            {
                "NormalizeImage": {
                    "std": [58.395, 57.12, 57.375],
                    "mean": [123.675, 116.28, 103.53],
                    "scale": "1",
                    "order": "hwc",
                }
            },
            {"ToCHWImage": None},
            {
                "KeepKeys": {
                    "keep_keys": [
                        "input_ids",
                        "bbox",
                        "attention_mask",
                        "token_type_ids",
                        "image",
                        "labels",
                        "segment_offset_id",
                        "ocr_info",
                        "entities",
                    ]
                }
            },
        ]

        self.preprocess_op = create_operators(pre_process_list, {"infer_mode": True})

    def _transform(self, data, ops=None):
        """transform"""
        if ops is None:
            ops = []
        for op in ops:
            data = op(data)
            if data is None:
                return None
        return data

    def run(self, input_im):
        """Run preprocess of  Ser-Vi-LayoutXLM model
        :param: input_ims: (numpy.ndarray) input image
        """
        ori_im = input_im.copy()
        data = {"image": input_im}
        data = transform(data, self.preprocess_op)

        for idx in range(len(data)):
            if isinstance(data[idx], np.ndarray):
                data[idx] = np.expand_dims(data[idx], axis=0)
            else:
                data[idx] = [data[idx]]

        return data


class StructureV2SERViLayoutXLMModelPostprocessor:
    def __init__(self, class_path):
        """Create a postprocessor for Ser-Vi-LayoutXLM model.
        :param: class_path: (string) class file path
        """
        self.postprocessor_op = VQASerTokenLayoutLMPostProcess(class_path)

    def run(self, preds, batch=None, *args, **kwargs):
        """Run postprocess of  Ser-Vi-LayoutXLM model.
        :param: preds: (list) results of inferring
        """
        return self.postprocessor_op(preds, batch, *args, **kwargs)


class StructureV2SERViLayoutXLMModel(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        ser_dict_path,
        class_path,
        config_file="",
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load SERViLayoutXLM model provided by PP-StructureV2.

        :param model_file: (str)Path of model file, e.g ./ser_vi_layout_xlm/model.pdmodel.
        :param params_file: (str)Path of parameter file, e.g ./ser_vi_layout_xlm/model.pdiparams, if the model format is ONNX, this parameter will be ignored.
        :param ser_dict_path: (str) class file path
        :param class_path: (str) class file path
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU.
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model.
        """
        super(StructureV2SERViLayoutXLMModel, self).__init__(runtime_option)

        assert (
            self._runtime_option.backend != 0
        ), "Runtime Option required backend setting."
        self._model = C.vision.ocr.StructureV2SERViLayoutXLMModel(
            model_file, params_file, config_file, self._runtime_option, model_format
        )

        assert self.initialized, "SERViLayoutXLM model initialize failed."

        self.preprocessor = StructureV2SERViLayoutXLMModelPreprocessor(ser_dict_path)
        self.postprocesser = StructureV2SERViLayoutXLMModelPostprocessor(class_path)

        self.input_name_0 = self._model.get_input_info(0).name
        self.input_name_1 = self._model.get_input_info(1).name
        self.input_name_2 = self._model.get_input_info(2).name
        self.input_name_3 = self._model.get_input_info(3).name

    def predict(self, image):
        assert isinstance(image, np.ndarray), "predict receives numpy.ndarray(BGR)"

        data = self.preprocessor.run(image)
        infer_input = {
            self.input_name_0: data[0],
            self.input_name_1: data[1],
            self.input_name_2: data[2],
            self.input_name_3: data[3],
        }

        infer_result = self._model.infer(infer_input)
        infer_result = infer_result[0]

        post_result = self.postprocesser.run(
            infer_result, segment_offset_ids=data[6], ocr_infos=data[7]
        )

        return post_result

    def batch_predict(self, image_list):
        assert isinstance(image_list, list) and isinstance(
            image_list[0], np.ndarray
        ), "batch_predict receives list of numpy.ndarray(BGR)"

        # reading and preprocessing images
        datas = None
        for image in image_list:
            data = self.preprocessor.run(image)

            # concatenate data to batch
            if datas == None:
                datas = data
            else:
                for idx in range(len(data)):
                    if isinstance(data[idx], np.ndarray):
                        datas[idx] = np.concatenate((datas[idx], data[idx]), axis=0)
                    else:
                        datas[idx].extend(data[idx])

        # infer
        infer_inputs = {
            self.input_name_0: datas[0],
            self.input_name_1: datas[1],
            self.input_name_2: datas[2],
            self.input_name_3: datas[3],
        }

        infer_results = self._model.infer(infer_inputs)
        infer_results = infer_results[0]

        # postprocessing
        post_results = self.postprocesser.run(
            infer_results, segment_offset_ids=datas[6], ocr_infos=datas[7]
        )

        return post_results


class PyOnlyFormulaRecognitionModel(PyOnlyVisionModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        self._model_file = model_file
        self._params_file = params_file
        self._model_format = model_format
        super().__init__(runtime_option)
        self._config = load_config(config_file)
        self._preprocessor = _PyOnlyFormulaRecognitionPreprocessor()
        self._postprocessor = _PyOnlyFormulaRecognitionPostprocessor(
            **self._config["PostProcess"]
        )

    def model_name():
        return "PyOnlyFormulaRecognitionModel"

    def batch_predict(self, imgs):
        data_list = []
        for img in imgs:
            data = {"img": img}
            data = self._preprocessor.run(data)
            data_list.append(data)

        input_name = self._runtime.get_input_info(0).name
        imgs = np.stack([data["img"] for data in data_list], axis=0, dtype=np.float32)
        imgs = np.ascontiguousarray(imgs)
        output_arrs = self._runtime.infer({input_name: imgs})

        results = []
        for score_map in output_arrs[0]:
            data = {"score_map": score_map}
            result = self._postprocessor.run(data)
            results.append(result)
        return results

    def _update_option(self):
        self._option.set_model_path(
            self._model_file, self._params_file, self._model_format
        )


class _PyOnlyFormulaRecognitionPreprocessor(object):
    def __init__(self):
        super().__init__()
        processors = self._build_processors()
        self._processor_chain = PyOnlyProcessorChain(processors)

    def run(self, data):
        return self._processor_chain(data)

    def _build_processors(self):
        processors = []
        processors.append(P.LaTeXOCRReisizeNormImg())
        return processors


class _PyOnlyFormulaRecognitionPostprocessor(object):
    def __init__(self, **kwargs):
        super().__init__()
        if kwargs.get("name") == "LaTeXOCRDecode":
            self.op = LaTeXOCRDecode(
                character_list=kwargs.get("character_dict"),
            )
        else:
            raise Exception()

    def run(self, data):
        rec_text = self.op.apply(data)
        rec_text = rec_text["rec_text"]
        result = _PyOnlyFormulaRecognitionResult(rec_text=rec_text)
        return result


@dataclass
class _PyOnlyFormulaRecognitionResult(object):
    rec_text: str


class LaTeXOCRDecode(object):
    def __init__(self, character_list=None):
        from tokenizers import Tokenizer as TokenizerFast

        super().__init__()
        character_list = character_list
        temp_path = tempfile.gettempdir()
        rec_char_dict_path = os.path.join(temp_path, "latexocr_tokenizer.json")
        try:
            with open(rec_char_dict_path, "w") as f:
                json.dump(character_list, f)
        except Exception as e:
            print(f"创建 latexocr_tokenizer.json 文件失败, 原因{str(e)}")
        self.tokenizer = TokenizerFast.from_file(rec_char_dict_path)

    def post_process(self, s):
        text_reg = r"(\\(operatorname|mathrm|text|mathbf)\s?\*? {.*?})"
        letter = "[a-zA-Z]"
        noletter = "[\W_^\d]"
        names = [x[0].replace(" ", "") for x in re.findall(text_reg, s)]
        s = re.sub(text_reg, lambda match: str(names.pop(0)), s)
        news = s
        while True:
            s = news
            news = re.sub(r"(?!\\ )(%s)\s+?(%s)" % (noletter, noletter), r"\1\2", s)
            news = re.sub(r"(?!\\ )(%s)\s+?(%s)" % (noletter, letter), r"\1\2", news)
            news = re.sub(r"(%s)\s+?(%s)" % (letter, noletter), r"\1\2", news)
            if news == s:
                break
        return s

    def decode(self, tokens):
        if len(tokens.shape) == 1:
            tokens = tokens[None, :]

        dec = [self.tokenizer.decode(tok) for tok in tokens]
        dec_str_list = [
            "".join(detok.split(" "))
            .replace("Ġ", " ")
            .replace("[EOS]", "")
            .replace("[BOS]", "")
            .replace("[PAD]", "")
            .strip()
            for detok in dec
        ]
        return [str(self.post_process(dec_str)) for dec_str in dec_str_list]

    def apply(self, pred):
        key = next(iter(pred))
        preds = np.array(pred[key])
        text = self.decode(preds)
        return {"rec_text": text[0]}



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
import cv2

from ..base import PyOnlyProcessor

__all__ = [
    "GetImageInfo",
    "Flip",
    "Crop",
    "Resize",
    "ResizeByLong",
    "ResizeByShort",
    "Pad",
    "PadStride",
    "Normalize",
    "ToCHWImage",
    "LaTeXOCRReisizeNormImg",
]


def _resize(im, target_size, interp):
    w, h = target_size
    im = cv2.resize(im, (w, h), interpolation=interp)
    return im


def _flip_h(im):
    if len(im.shape) == 3:
        im = im[:, ::-1, :]
    elif len(im.shape) == 2:
        im = im[:, ::-1]
    return im


def _flip_v(im):
    if len(im.shape) == 3:
        im = im[::-1, :, :]
    elif len(im.shape) == 2:
        im = im[::-1, :]
    return im


def _slice(im, coords):
    x1, y1, x2, y2 = coords
    im = im[y1:y2, x1:x2, ...]
    return im


def _pad(im, pad, val):
    if isinstance(pad, int):
        pad = [pad] * 4
    if len(pad) != 4:
        raise ValueError
    chns = 1 if im.ndim == 2 else im.shape[2]
    im = cv2.copyMakeBorder(im, *pad, cv2.BORDER_CONSTANT, value=(val,) * chns)
    return im


def _check_image_size(input_):
    if not (
        isinstance(input_, (list, tuple))
        and len(input_) == 2
        and isinstance(input_[0], int)
        and isinstance(input_[1], int)
    ):
        raise TypeError(f"{input_} cannot represent a valid image size.")


class GetImageInfo(PyOnlyProcessor):
    def __call__(self, data):
        img = data["img"]

        return {**data, "img_size": [img.shape[1], img.shape[0]]}


class Flip(PyOnlyProcessor):
    def __init__(self, mode="H"):
        super().__init__()
        if mode not in ("H", "V"):
            raise ValueError("`mode` should be 'H' or 'V'.")
        self._mode = mode

    def __call__(self, data):
        img = data["img"]

        if self._mode == "H":
            img = _flip_h(img)
        elif self._mode == "V":
            img = _flip_v(img)

        return {**data, "img": img}


class Crop(PyOnlyProcessor):
    def __init__(self, crop_size, mode="C"):
        super().__init__()
        if isinstance(crop_size, int):
            crop_size = [crop_size, crop_size]
        _check_image_size(crop_size)

        self._crop_size = crop_size

        if mode not in ("C", "TL"):
            raise ValueError("Unsupported interpolation method")
        self._mode = mode

    def __call__(self, data):
        img = data["img"]

        h, w = img.shape[:2]
        cw, ch = self._crop_size
        if self._mode == "C":
            x1 = max(0, (w - cw) // 2)
            y1 = max(0, (h - ch) // 2)
        elif self._mode == "TL":
            x1, y1 = 0, 0
        x2 = min(w, x1 + cw)
        y2 = min(h, y1 + ch)
        coords = (x1, y1, x2, y2)
        if coords == (0, 0, w, h):
            raise ValueError(
                f"Input image ({w}, {h}) smaller than the target size ({cw}, {ch})."
            )
        img = _slice(img, coords=coords)

        return {**data, "img": img, "img_size": [img.shape[1], img.shape[0]]}


class _BaseResize(PyOnlyProcessor):
    _INTERP_DICT = {
        "NEAREST": cv2.INTER_NEAREST,
        "LINEAR": cv2.INTER_LINEAR,
        "CUBIC": cv2.INTER_CUBIC,
        "AREA": cv2.INTER_AREA,
        "LANCZOS4": cv2.INTER_LANCZOS4,
    }

    def __init__(self, size_divisor, interp):
        super().__init__()

        if size_divisor is not None:
            assert isinstance(
                size_divisor, int
            ), "`size_divisor` should be None or int."
        self._size_divisor = size_divisor

        try:
            interp = self._INTERP_DICT[interp]
        except KeyError:
            raise ValueError(
                "`interp` should be one of {}.".format(self._INTERP_DICT.keys())
            )
        self._interp = interp

    @staticmethod
    def _rescale_size(img_size, target_size):
        scale = min(max(target_size) / max(img_size), min(target_size) / min(img_size))
        rescaled_size = [round(i * scale) for i in img_size]
        return rescaled_size, scale


class Resize(_BaseResize):
    def __init__(
        self, target_size, keep_ratio=False, size_divisor=None, interp="LINEAR"
    ):
        super().__init__(size_divisor=size_divisor, interp=interp)

        if isinstance(target_size, int):
            target_size = [target_size, target_size]
        _check_image_size(target_size)
        self._target_size = target_size

        self._keep_ratio = keep_ratio

    def __call__(self, data):
        img = data["img"]

        target_size = self._target_size
        original_size = img.shape[:2][::-1]

        if self._keep_ratio:
            h, w = img.shape[0:2]
            target_size, _ = self._rescale_size((w, h), self._target_size)

        if self._size_divisor:
            target_size = [
                math.ceil(i / self._size_divisor) * self._size_divisor
                for i in target_size
            ]

        img_scale_w, img_scale_h = [
            target_size[0] / original_size[0],
            target_size[1] / original_size[1],
        ]
        img = _resize(img, target_size, interp=self._interp)

        return {
            **data,
            "img": img,
            "img_size": [img.shape[1], img.shape[0]],
            "scale_factors": [img_scale_w, img_scale_h],
        }


class ResizeByLong(_BaseResize):
    def __init__(self, target_long_edge, size_divisor=None, interp="LINEAR"):
        super().__init__(size_divisor=size_divisor, interp=interp)
        self._target_long_edge = target_long_edge

    def __call__(self, data):
        img = data["img"]

        h, w = img.shape[:2]
        scale = self._target_long_edge / max(h, w)
        h_resize = round(h * scale)
        w_resize = round(w * scale)
        if self._size_divisor is not None:
            h_resize = math.ceil(h_resize / self._size_divisor) * self._size_divisor
            w_resize = math.ceil(w_resize / self._size_divisor) * self._size_divisor

        img = _resize(img, (w_resize, h_resize), interp=self._interp)

        return {**data, "img": img, "img_size": [img.shape[1], img.shape[0]]}


class ResizeByShort(_BaseResize):
    INPUT_KEYS = "img"
    OUTPUT_KEYS = ["img", "img_size"]
    DEAULT_INPUTS = {"img": "img"}
    DEAULT_OUTPUTS = {"img": "img", "img_size": "img_size"}

    def __init__(self, target_short_edge, size_divisor=None, interp="LINEAR"):
        super().__init__(size_divisor=size_divisor, interp=interp)
        self._target_short_edge = target_short_edge

    def __call__(self, data):
        img = data["img"]

        h, w = img.shape[:2]
        scale = self._target_short_edge / min(h, w)
        h_resize = round(h * scale)
        w_resize = round(w * scale)
        if self._size_divisor is not None:
            h_resize = math.ceil(h_resize / self._size_divisor) * self._size_divisor
            w_resize = math.ceil(w_resize / self._size_divisor) * self._size_divisor

        img = _resize(img, (w_resize, h_resize), interp=self._interp)

        return {**data, "img": img, "img_size": [img.shape[1], img.shape[0]]}


class Pad(PyOnlyProcessor):
    def __init__(self, target_size, val=127.5):
        super().__init__()

        if isinstance(target_size, int):
            target_size = [target_size, target_size]
        _check_image_size(target_size)
        self._target_size = target_size

        self._val = val

    def __call__(self, data):
        img = data["img"]

        h, w = img.shape[:2]
        tw, th = self._target_size
        ph = th - h
        pw = tw - w

        if ph < 0 or pw < 0:
            raise ValueError(
                f"Input image ({w}, {h}) smaller than the target size ({tw}, {th})."
            )
        else:
            img = _pad(img, pad=(0, ph, 0, pw), val=self._val)

        return {**data, "img": img, "img_size": [img.shape[1], img.shape[0]]}


class PadStride(PyOnlyProcessor):
    INPUT_KEYS = "img"
    OUTPUT_KEYS = "img"
    DEAULT_INPUTS = {"img": "img"}
    DEAULT_OUTPUTS = {"img": "img"}

    def __init__(self, stride=0):
        super().__init__()
        self._coarsest_stride = stride

    def __call__(self, data):
        img = data["img"]

        im = img
        coarsest_stride = self._coarsest_stride
        if coarsest_stride <= 0:
            return {"img": im}
        im_c, im_h, im_w = im.shape
        pad_h = int(np.ceil(float(im_h) / coarsest_stride) * coarsest_stride)
        pad_w = int(np.ceil(float(im_w) / coarsest_stride) * coarsest_stride)
        padding_im = np.zeros((im_c, pad_h, pad_w), dtype=np.float32)
        padding_im[:, :im_h, :im_w] = im

        return {**data, "img": padding_im}


class Normalize(PyOnlyProcessor):
    def __init__(self, scale=1.0 / 255, mean=0.5, std=0.5, preserve_dtype=False):
        super().__init__()
        self._scale = np.float32(scale)
        if isinstance(mean, float):
            mean = [mean]
        self._mean = np.asarray(mean).astype("float32")
        if isinstance(std, float):
            std = [std]
        self._std = np.asarray(std).astype("float32")
        self._preserve_dtype = preserve_dtype

    def __call__(self, data):
        img = data["img"]

        old_type = img.dtype
        # XXX: If `old_type` has higher precision than float32,
        # we will lose some precision.
        img = img.astype("float32", copy=False)
        img *= self._scale
        img -= self._mean
        img /= self._std
        if self._preserve_dtype:
            img = img.astype(old_type, copy=False)

        return {**data, "img": img}


class ToCHWImage(PyOnlyProcessor):
    def __call__(self, data):
        img = data["img"]

        img = img.transpose((2, 0, 1))

        return {**data, "img": img}


class BGR2RGB(PyOnlyProcessor):
    def __call__(self, data):
        img = data["img"]

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return {**data, "img": img}


class LaTeXOCRReisizeNormImg(PyOnlyProcessor):
    """for ocr image resize and normalization"""

    def __init__(self, rec_image_shape=(3, 48, 320)):
        super().__init__()
        self.rec_image_shape = rec_image_shape

    def pad_(self, img, divable=32):
        from PIL import Image

        threshold = 128
        data = np.array(img.convert("LA"))
        if data[..., -1].var() == 0:
            data = (data[..., 0]).astype(np.uint8)
        else:
            data = (255 - data[..., -1]).astype(np.uint8)
        data = (data - data.min()) / (data.max() - data.min()) * 255
        if data.mean() > threshold:
            # To invert the text to white
            gray = 255 * (data < threshold).astype(np.uint8)
        else:
            gray = 255 * (data > threshold).astype(np.uint8)
            data = 255 - data

        coords = cv2.findNonZero(gray)  # Find all non-zero points (text)
        a, b, w, h = cv2.boundingRect(coords)  # Find minimum spanning bounding box
        rect = data[b : b + h, a : a + w]
        im = Image.fromarray(rect).convert("L")
        dims = []
        for x in [w, h]:
            div, mod = divmod(x, divable)
            dims.append(divable * (div + (1 if mod > 0 else 0)))
        padded = Image.new("L", dims, 255)
        padded.paste(im, (0, 0, im.size[0], im.size[1]))
        return padded

    def minmax_size_(
        self,
        img,
        max_dimensions,
        min_dimensions,
    ):
        from PIL import Image

        if max_dimensions is not None:
            ratios = [a / b for a, b in zip(img.size, max_dimensions)]
            if any([r > 1 for r in ratios]):
                size = np.array(img.size) // max(ratios)
                img = img.resize(tuple(size.astype(int)), Image.BILINEAR)
        if min_dimensions is not None:
            # hypothesis: there is a dim in img smaller than min_dimensions, and return a proper dim >= min_dimensions
            padded_size = [
                max(img_dim, min_dim)
                for img_dim, min_dim in zip(img.size, min_dimensions)
            ]
            if padded_size != list(img.size):  # assert hypothesis
                padded_im = Image.new("L", padded_size, 255)
                padded_im.paste(img, img.getbbox())
                img = padded_im
        return img

    def norm_img_latexocr(self, img):
        # CAN only predict gray scale image
        from PIL import Image

        shape = (1, 1, 3)
        mean = [0.7931, 0.7931, 0.7931]
        std = [0.1738, 0.1738, 0.1738]
        scale = np.float32(1.0 / 255.0)
        min_dimensions = [32, 32]
        max_dimensions = [672, 192]
        mean = np.array(mean).reshape(shape).astype("float32")
        std = np.array(std).reshape(shape).astype("float32")

        im_h, im_w = img.shape[:2]
        if (
            min_dimensions[0] <= im_w <= max_dimensions[0]
            and min_dimensions[1] <= im_h <= max_dimensions[1]
        ):
            pass
        else:
            img = Image.fromarray(np.uint8(img))
            img = self.minmax_size_(self.pad_(img), max_dimensions, min_dimensions)
            img = np.array(img)
            im_h, im_w = img.shape[:2]
            img = np.dstack([img, img, img])
        img = (img.astype("float32") * scale - mean) / std
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        divide_h = math.ceil(im_h / 16) * 16
        divide_w = math.ceil(im_w / 16) * 16
        img = np.pad(
            img, ((0, divide_h - im_h), (0, divide_w - im_w)), constant_values=(1, 1)
        )
        img = img[:, :, np.newaxis].transpose(2, 0, 1)
        img = img.astype("float32")
        return img

    def __call__(self, data):
        """apply"""
        img = data["img"]
        img = self.norm_img_latexocr(img)
        return {"img": img}



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
import io
import os
import warnings
from pathlib import Path
from typing import List, Optional

import numpy as np

from ....modules.doc_vlm.model_list import MODELS
from ....utils import logging
from ....utils.deps import require_genai_client_plugin
from ....utils.device import TemporaryDeviceChanger
from ...common.batch_sampler import DocVLMBatchSampler
from ...utils.misc import is_bfloat16_available
from ...utils.model_paths import get_model_paths
from ..base import BasePredictor
from .result import DocVLMResult


class DocVLMPredictor(BasePredictor):

    entities = MODELS
    model_group = {
        "PP-DocBee": {"PP-DocBee-2B", "PP-DocBee-7B"},
        "PP-DocBee2": {"PP-DocBee2-3B"},
        "PP-Chart2Table": {"PP-Chart2Table"},
        "PaddleOCR-VL": {"PaddleOCR-VL-0.9B", "PaddleOCR-VL-1.5-0.9B"},
    }

    def __init__(self, *args, **kwargs):
        """Initializes DocVLMPredictor.
        Args:
            *args: Arbitrary positional arguments passed to the superclass.
            **kwargs: Arbitrary keyword arguments passed to the superclass.
        """
        super().__init__(*args, **kwargs)

        if self.batch_sampler.batch_size == -1:
            self.batch_sampler.batch_size = self._determine_batch_size()

        if self._use_local_model:
            if self._use_static_model:
                raise RuntimeError("Static graph models are not supported")
            self.device = kwargs.get("device", None)
            if is_bfloat16_available(self.device):
                self.dtype = "bfloat16"
            else:
                self.dtype = "float32"

            self.infer, self.processor = self._build(**kwargs)

            if (
                self.model_name in self.model_group["PaddleOCR-VL"]
                and self.batch_sampler.batch_size > 1
            ):
                logging.warning(
                    f"Currently, the {repr(self.model_name)} local model only supports batch size of 1. The batch size will be updated to 1."
                )
                self.batch_sampler.batch_size = 1

    def _build_batch_sampler(self):
        """Builds and returns an DocVLMBatchSampler instance.

        Returns:
            DocVLMBatchSampler: An instance of DocVLMBatchSampler.
        """
        return DocVLMBatchSampler(self.model_name)

    def _get_result_class(self):
        """Returns the result class, DocVLMResult.

        Returns:
            type: The DocVLMResult class.
        """
        return DocVLMResult

    def _build(self, **kwargs):
        """Build the model, and correspounding processor on the configuration.

        Returns:
            model: An instance of Paddle model, could be either a dynamic model or a static model.
            processor: The correspounding processor for the model.
        """
        from .modeling import (
            PaddleOCRVLForConditionalGeneration,
            PPChart2TableInference,
            PPDocBee2Inference,
            PPDocBeeInference,
        )

        # build processor
        processor = self.build_processor()

        # build model
        if self.model_name in self.model_group["PP-DocBee"]:
            if kwargs.get("use_hpip", False):
                warnings.warn(
                    "The PP-DocBee series does not support `use_hpip=True` for now."
                )
            with TemporaryDeviceChanger(self.device):
                model = PPDocBeeInference.from_pretrained(
                    self.model_dir, dtype=self.dtype
                )
        elif self.model_name in self.model_group["PP-Chart2Table"]:
            if kwargs.get("use_hpip", False):
                warnings.warn(
                    "The PP-Chart2Table series does not support `use_hpip=True` for now."
                )
            with TemporaryDeviceChanger(self.device):
                model_path = get_model_paths(self.model_dir)

                if "safetensors" in model_path:
                    model = PPChart2TableInference.from_pretrained(
                        self.model_dir,
                        dtype=self.dtype,
                        pad_token_id=processor.tokenizer.eos_token_id,
                        use_safetensors=True,
                        convert_from_hf=True,
                    )
                else:
                    model = PPChart2TableInference.from_pretrained(
                        self.model_dir,
                        dtype=self.dtype,
                        pad_token_id=processor.tokenizer.eos_token_id,
                    )

        elif self.model_name in self.model_group["PP-DocBee2"]:
            if kwargs.get("use_hpip", False):
                warnings.warn(
                    "The PP-Chart2Table series does not support `use_hpip=True` for now."
                )
            with TemporaryDeviceChanger(self.device):
                model = PPDocBee2Inference.from_pretrained(
                    self.model_dir,
                    dtype=self.dtype,
                )
        elif self.model_name in self.model_group["PaddleOCR-VL"]:
            if kwargs.get("use_hpip", False):
                warnings.warn(
                    "The PaddelOCR-VL series does not support `use_hpip=True` for now."
                )
            with TemporaryDeviceChanger(self.device):
                model = PaddleOCRVLForConditionalGeneration.from_pretrained(
                    self.model_dir,
                    dtype=self.dtype,
                    convert_from_hf=True,
                )
        else:
            raise NotImplementedError(f"Model {self.model_name} is not supported.")

        return model, processor

    def _determine_batch_size(self):
        if self._model_name == "PaddleOCR-VL-0.9B":
            batch_size = 1
            if not self._use_local_model:
                batch_size = 4096
            logging.debug(
                f"The batch size of {self._model_name} is determined to be {batch_size}."
            )
            return batch_size
        else:
            raise RuntimeError(f"Could not determine batch size for {self._model_name}")

    def process(
        self,
        data: List[dict],
        max_new_tokens: Optional[int] = None,
        skip_special_tokens: Optional[bool] = None,
        repetition_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        min_pixels: Optional[int] = None,
        max_pixels: Optional[int] = None,
        use_cache: Optional[bool] = None,
        **kwargs,
    ):
        """
        Process a batch of data through the preprocessing, inference, and postprocessing.

        Args:
            data (List[dict]): A batch of input data, must be a dict (e.g. {"image": /path/to/image, "query": some question}).

        Returns:
            dict: A dictionary containing the raw sample information and prediction results for every instance of the batch.
        """
        # TODO: Sampling settings
        # FIXME: When `skip_special_tokens` is `True`, the results from different backends may differ.

        assert all(isinstance(i, dict) for i in data)

        if self._use_local_model:
            src_data = copy.copy(data)
            # preprocess
            if self.model_name in self.model_group["PaddleOCR-VL"]:
                data = self.processor.preprocess(
                    data, min_pixels=min_pixels, max_pixels=max_pixels
                )
            else:
                data = self.processor.preprocess(data)
                if min_pixels is not None:
                    warnings.warn(
                        f"`min_pixels` is currently not supported by the {repr(self.model_name)} model and will be ignored."
                    )
                if max_pixels is not None:
                    warnings.warn(
                        f"`max_pixels` is currently not supported by the {repr(self.model_name)} model and will be ignored."
                    )

            data = self._switch_inputs_to_device(data)

            # do infer
            generate_kwargs = {}
            if max_new_tokens is not None:
                generate_kwargs["max_new_tokens"] = max_new_tokens
            elif self.model_name in self.model_group["PaddleOCR-VL"]:
                generate_kwargs["max_new_tokens"] = 8192
            if repetition_penalty is not None:
                warnings.warn(
                    "`repetition_penalty` is currently not supported by the local model and will be ignored."
                )
            if temperature is not None:
                warnings.warn(
                    "`temperature` is currently not supported by the local model and will be ignored."
                )
            if top_p is not None:
                warnings.warn(
                    "`top_p` is currently not supported by the local model and will be ignored."
                )
            if use_cache is not None:
                generate_kwargs["use_cache"] = use_cache
            with TemporaryDeviceChanger(self.device):
                preds = self.infer.generate(
                    data,
                    **generate_kwargs,
                )

            # postprocess
            postprocess_kwargs = {}
            if skip_special_tokens is not None:
                postprocess_kwargs["skip_special_tokens"] = skip_special_tokens
            preds = self.processor.postprocess(preds, **postprocess_kwargs)
        else:
            require_genai_client_plugin()

            src_data = data

            preds = self._genai_client_process(
                data,
                max_new_tokens=max_new_tokens,
                skip_special_tokens=skip_special_tokens,
                repetition_penalty=repetition_penalty,
                temperature=temperature,
                top_p=top_p,
                min_pixels=min_pixels,
                max_pixels=max_pixels,
            )

        result_dict = self._format_result_dict(preds, src_data)
        return result_dict

    def build_processor(self, **kwargs):
        from ..common.tokenizer import (
            LlamaTokenizer,
            MIXQwen2_5_Tokenizer,
            MIXQwen2Tokenizer,
            QWenTokenizer,
        )
        from ..common.tokenizer.tokenizer_utils import ChatTemplate
        from .processors import (
            GOTImageProcessor,
            PaddleOCRVLProcessor,
            PPChart2TableProcessor,
            PPDocBee2Processor,
            PPDocBeeProcessor,
            Qwen2_5_VLImageProcessor,
            Qwen2VLImageProcessor,
            SiglipImageProcessor,
        )

        if self.model_name in self.model_group["PP-DocBee"]:
            image_processor = Qwen2VLImageProcessor()
            tokenizer = MIXQwen2Tokenizer.from_pretrained(self.model_dir)
            return PPDocBeeProcessor(
                image_processor=image_processor, tokenizer=tokenizer
            )
        elif self.model_name in self.model_group["PP-Chart2Table"]:
            image_processor = GOTImageProcessor(1024)
            tokenizer = QWenTokenizer.from_pretrained(self.model_dir)
            return PPChart2TableProcessor(
                image_processor=image_processor, tokenizer=tokenizer, dtype=self.dtype
            )
        elif self.model_name in self.model_group["PP-DocBee2"]:
            image_processor = Qwen2_5_VLImageProcessor()
            tokenizer = MIXQwen2_5_Tokenizer.from_pretrained(self.model_dir)
            return PPDocBee2Processor(
                image_processor=image_processor, tokenizer=tokenizer
            )
        elif self.model_name in self.model_group["PaddleOCR-VL"]:
            image_processor = SiglipImageProcessor.from_pretrained(self.model_dir)
            vocab_file = str(Path(self.model_dir, "tokenizer.model"))
            tokenizer = LlamaTokenizer.from_pretrained(
                self.model_dir, vocab_file=vocab_file
            )
            # HACK
            chat_template_file = Path(self.model_dir, "chat_template.jinja")
            tokenizer.chat_template = ChatTemplate._compile_jinja_template(
                chat_template_file.read_text(encoding="utf-8")
            )
            return PaddleOCRVLProcessor(
                image_processor=image_processor,
                tokenizer=tokenizer,
            )
        else:
            raise NotImplementedError

    def _format_result_dict(self, model_preds, src_data):
        if not isinstance(model_preds, list):
            model_preds = [model_preds]
        if not isinstance(src_data, list):
            src_data = [src_data]
        input_info = []
        for data in src_data:
            image = data.get("image", None)
            if isinstance(image, str):
                data["input_path"] = image
            input_info.append(data)
        if len(model_preds) != len(input_info):
            raise ValueError(
                f"Model predicts {len(model_preds)} results while src data has {len(input_info)} samples."
            )

        rst_format_dict = {k: [] for k in input_info[0].keys()}
        rst_format_dict["result"] = []

        for data_sample, model_pred in zip(input_info, model_preds):
            for k in data_sample.keys():
                rst_format_dict[k].append(data_sample[k])
            rst_format_dict["result"].append(model_pred)

        return rst_format_dict

    def _infer_dynamic_forward_device(self, device):
        """infer the forward device for dynamic graph model"""
        import GPUtil

        from ....utils.device import parse_device

        if device is None:
            return None
        if "cpu" in device.lower():
            return "cpu"
        device_type, device_ids = parse_device(device)

        cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
        if cuda_visible_devices is None:
            env_gpu_num = len(GPUtil.getGPUs())
            cuda_visible_devices = ",".join([str(i) for i in range(env_gpu_num)])
        env_device_ids = cuda_visible_devices.split(",")
        for env_device_id in env_device_ids:
            if not env_device_id.isdigit():
                raise ValueError(
                    f"CUDA_VISIBLE_DEVICES ID must be an integer. Invalid device ID: {env_device_id}"
                )

        if max(device_ids) >= len(env_device_ids):
            raise ValueError(
                f"Required gpu ids {device_ids} even larger than the number of visible devices {cuda_visible_devices}."
            )

        rst_global_gpu_ids = [env_device_ids[idx] for idx in device_ids]
        return device_type + ":" + ",".join(rst_global_gpu_ids)

    def _switch_inputs_to_device(self, input_dict):
        """Switch the input to the specified device"""
        import paddle

        if self.device is None:
            return input_dict
        rst_dict = {
            k: (
                paddle.to_tensor(input_dict[k], place=self.device)
                if isinstance(input_dict[k], paddle.Tensor)
                else input_dict[k]
            )
            for k in input_dict
        }
        return rst_dict

    def _genai_client_process(
        self,
        data,
        max_new_tokens,
        skip_special_tokens,
        repetition_penalty,
        temperature,
        top_p,
        min_pixels,
        max_pixels,
    ):
        futures = []
        for item in data:
            image = item["image"]
            if isinstance(image, str):
                if image.startswith("http://") or image.startswith("https://"):
                    image_url = image
                else:
                    from PIL import Image

                    with Image.open(image) as img:
                        img = img.convert("RGB")
                        with io.BytesIO() as buf:
                            img.save(buf, format="JPEG")
                            image_url = "data:image/jpeg;base64," + base64.b64encode(
                                buf.getvalue()
                            ).decode("ascii")
            elif isinstance(image, np.ndarray):
                import cv2
                from PIL import Image

                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                img = Image.fromarray(image)
                with io.BytesIO() as buf:
                    img.save(buf, format="JPEG")
                    image_url = "data:image/jpeg;base64," + base64.b64encode(
                        buf.getvalue()
                    ).decode("ascii")
            else:
                raise TypeError(f"Not supported image type: {type(image)}")

            if self._genai_client.backend == "fastdeploy-server":
                kwargs = {
                    "temperature": 1 if temperature is None else temperature,
                    "top_p": 0 if top_p is None else top_p,
                }
            else:
                kwargs = {
                    "temperature": 0 if temperature is None else temperature,
                }
                if top_p is not None:
                    kwargs["top_p"] = top_p

            if self._genai_client.backend == "mlx-vlm-server":
                max_tokens_name = "max_tokens"
            else:
                max_tokens_name = "max_completion_tokens"

            if max_new_tokens is not None:
                kwargs[max_tokens_name] = max_new_tokens
            elif self.model_name in self.model_group["PaddleOCR-VL"]:
                kwargs[max_tokens_name] = 8192

            kwargs["extra_body"] = {}
            if skip_special_tokens is not None:
                if self._genai_client.backend in (
                    "fastdeploy-server",
                    "vllm-server",
                    "sglang-server",
                    "mlx-vlm-server",
                ):
                    kwargs["extra_body"]["skip_special_tokens"] = skip_special_tokens
                else:
                    raise ValueError("Not supported")

            if repetition_penalty is not None:
                kwargs["extra_body"]["repetition_penalty"] = repetition_penalty

            if min_pixels is not None:
                if self._genai_client.backend == "vllm-server":
                    kwargs["extra_body"]["mm_processor_kwargs"] = kwargs[
                        "extra_body"
                    ].get("mm_processor_kwargs", {})
                    kwargs["extra_body"]["mm_processor_kwargs"][
                        "min_pixels"
                    ] = min_pixels
                else:
                    warnings.warn(
                        f"{repr(self._genai_client.backend)} does not support `min_pixels`."
                    )

            if max_pixels is not None:
                if self._genai_client.backend == "vllm-server":
                    kwargs["extra_body"]["mm_processor_kwargs"] = kwargs[
                        "extra_body"
                    ].get("mm_processor_kwargs", {})
                    kwargs["extra_body"]["mm_processor_kwargs"][
                        "max_pixels"
                    ] = max_pixels
                else:
                    warnings.warn(
                        f"{repr(self._genai_client.backend)} does not support `max_pixels`."
                    )

            future = self._genai_client.create_chat_completion(
                [
                    {
                        "role": "user",
                        "content": [
                            {"type": "image_url", "image_url": {"url": image_url}},
                            {"type": "text", "text": item["query"]},
                        ],
                    }
                ],
                return_future=True,
                timeout=600,
                **kwargs,
            )

            futures.append(future)

        results = []
        for future in futures:
            result = future.result()
            results.append(result.choices[0].message.content)

        return results



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import mimetypes
import shutil
from urllib.parse import urlparse
from urllib.request import urlopen


def is_url(s):
    if not (s.startswith("http://") or s.startswith("https://")):
        # Quick rejection
        return False
    result = urlparse(s)
    return all([result.scheme, result.netloc]) and result.scheme in ("http", "https")


def prepare_input_file(file, include_header=False):
    if is_url(file):
        return file
    else:
        with open(file, "rb") as f:
            bytes_ = f.read()
        encoded = base64.b64encode(bytes_).decode("ascii")
        if include_header:
            mime_type = mimetypes.guess_type(file)[0] or "application/octet-stream"
            return f"data:{mime_type};base64,{encoded}"
        return encoded


def save_output_file(file, path, include_header=False):
    if is_url(file):
        with urlopen(file) as r:
            with open(path, "wb") as f:
                shutil.copyfileobj(r, f)
    else:
        if include_header:
            header, encoded = file.split(",", 1)
            if not (header.startswith("data:") and header.endswith(";base64")):
                raise ValueError("Invalid data URI format")
        else:
            encoded = file
        bytes_ = base64.b64decode(encoded)
        with open(path, "wb") as f:
            f.write(bytes_)



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils.device import get_default_device, parse_device
from ...utils.env import get_device_type


def is_mkldnn_available():
    # XXX: Not sure if this is the best way to check if MKL-DNN is available
    from paddle.inference import Config

    return hasattr(Config, "set_mkldnn_cache_capacity")


def is_bfloat16_available(device):
    import paddle.amp

    if device is None:
        device = get_default_device()
    device_type, _ = parse_device(device)
    return (
        "npu" in get_device_type() or paddle.amp.is_bfloat16_supported()
    ) and device_type in ("gpu", "npu", "xpu", "mlu", "metax_gpu", "iluvatar_gpu")


def is_float16_available(device):
    import paddle.amp

    if device is None:
        device = get_default_device()
    device_type, _ = parse_device(device)
    return (
        "npu" in get_device_type() or paddle.amp.is_float16_supported()
    ) and device_type in (
        "gpu",
        "npu",
        "xpu",
        "mlu",
        "dcu",
        "metax_gpu",
        "iluvatar_gpu",
    )



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from concurrent.futures import ThreadPoolExecutor
from operator import itemgetter
from typing import Any, Dict, Final, List, Tuple

from paddlex_hps_server import (
    BaseTritonPythonModel,
    app_common,
    logging,
    protocol,
    schemas,
    utils,
)
from paddlex_hps_server.storage import SupportsGetURL, create_storage

_DEFAULT_MAX_NUM_INPUT_IMGS: Final[int] = 10
_DEFAULT_MAX_OUTPUT_IMG_SIZE: Final[Tuple[int, int]] = (2000, 2000)


class _SequentialExecutor(object):
    def map(self, fn, *iterables):
        return map(fn, *iterables)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class TritonPythonModel(BaseTritonPythonModel):
    def initialize(self, args):
        super().initialize(args)
        self.context = {}
        self.context["file_storage"] = None
        self.context["return_img_urls"] = False
        self.context["max_num_input_imgs"] = _DEFAULT_MAX_NUM_INPUT_IMGS
        self.context["max_output_img_size"] = _DEFAULT_MAX_OUTPUT_IMG_SIZE
        if self.app_config.extra:
            if "file_storage" in self.app_config.extra:
                self.context["file_storage"] = create_storage(
                    self.app_config.extra["file_storage"]
                )
            if "return_img_urls" in self.app_config.extra:
                self.context["return_img_urls"] = self.app_config.extra[
                    "return_img_urls"
                ]
            if "max_num_input_imgs" in self.app_config.extra:
                self.context["max_num_input_imgs"] = self.app_config.extra[
                    "max_num_input_imgs"
                ]
            if "max_output_img_size" in self.app_config.extra:
                self.context["max_output_img_size"] = self.app_config.extra[
                    "max_output_img_size"
                ]
        if self.context["return_img_urls"]:
            file_storage = self.context["file_storage"]
            if not file_storage:
                raise ValueError(
                    "The file storage must be properly configured when URLs need to be returned."
                )
            if not isinstance(file_storage, SupportsGetURL):
                raise TypeError(f"{type(file_storage)} does not support getting URLs.")

    def get_input_model_type(self):
        return schemas.ocr.InferRequest

    def get_result_model_type(self):
        return schemas.ocr.InferResult

    def run(self, input, log_id):
        return self.run_batch([input], [log_id], log_id)

    def run_batch(self, inputs, log_ids, batch_id):
        result_or_output_dic = {}

        input_groups = self._group_inputs(inputs)

        max_group_size = max(map(len, input_groups))
        if max_group_size > 1:
            executor = ThreadPoolExecutor(max_workers=max_group_size)
        else:
            executor = _SequentialExecutor()

        with executor:
            for input_group in input_groups:
                input_ids_g = list(map(itemgetter(0), input_group))
                inputs_g = list(map(itemgetter(1), input_group))

                log_ids_g = [log_ids[i] for i in input_ids_g]

                ret = executor.map(self._preprocess, inputs_g, log_ids_g)
                ind_img_lsts, ind_data_info_lst, ind_visualize_enabled_lst = [], [], []
                for i, item in enumerate(ret):
                    if isinstance(item, tuple):
                        assert len(item) == 3, len(item)
                        ind_img_lsts.append(item[0])
                        ind_data_info_lst.append(item[1])
                        ind_visualize_enabled_lst.append(item[2])
                    else:
                        input_id = input_ids_g[i]
                        result_or_output_dic[input_id] = item

                if len(ind_img_lsts):
                    images = [img for item in ind_img_lsts for img in item]
                    preds = list(
                        self.pipeline(
                            images,
                            use_doc_orientation_classify=inputs_g[
                                0
                            ].useDocOrientationClassify,
                            use_doc_unwarping=inputs_g[0].useDocUnwarping,
                            use_textline_orientation=inputs_g[0].useTextlineOrientation,
                            text_det_limit_side_len=inputs_g[0].textDetLimitSideLen,
                            text_det_limit_type=inputs_g[0].textDetLimitType,
                            text_det_thresh=inputs_g[0].textDetThresh,
                            text_det_box_thresh=inputs_g[0].textDetBoxThresh,
                            text_det_unclip_ratio=inputs_g[0].textDetUnclipRatio,
                            text_rec_score_thresh=inputs_g[0].textRecScoreThresh,
                            return_word_box=inputs_g[0].returnWordBox,
                        )
                    )

                    if len(preds) != len(images):
                        raise RuntimeError(
                            f"The number of predictions ({len(preds)}) is not the same as the number of input images ({len(images)})."
                        )

                    start_idx = 0
                    ind_preds = []
                    for item in ind_img_lsts:
                        ind_preds.append(preds[start_idx : start_idx + len(item)])
                        start_idx += len(item)

                    for i, result in zip(
                        input_ids_g,
                        executor.map(
                            self._postprocess,
                            ind_img_lsts,
                            ind_data_info_lst,
                            ind_visualize_enabled_lst,
                            ind_preds,
                            log_ids_g,
                            inputs_g,
                        ),
                    ):
                        result_or_output_dic[i] = result

            assert len(result_or_output_dic) == len(
                inputs
            ), f"Expected {len(inputs)} results or outputs, but got {len(result_or_output_dic)}"

            return [result_or_output_dic[i] for i in range(len(inputs))]

    def _group_inputs(self, inputs):
        def _to_hashable(obj):
            if isinstance(obj, dict):
                return tuple(
                    (_to_hashable(k), _to_hashable(v))
                    for k, v in sorted(obj.items(), key=lambda x: repr(x[0]))
                )
            elif isinstance(obj, list):
                return tuple(_to_hashable(x) for x in obj)
            else:
                return obj

        def _hash(input):
            return hash(
                tuple(
                    map(
                        _to_hashable,
                        (
                            input.useDocOrientationClassify,
                            input.useDocUnwarping,
                            input.useTextlineOrientation,
                            input.textDetLimitSideLen,
                            input.textDetLimitType,
                            input.textDetThresh,
                            input.textDetBoxThresh,
                            input.textDetUnclipRatio,
                            input.textRecScoreThresh,
                            input.returnWordBox,
                        ),
                    )
                )
            )

        groups = {}
        for i, inp in enumerate(inputs):
            group_key = _hash(inp)
            if group_key not in groups:
                groups[group_key] = []
            groups[group_key].append((i, inp))

        return list(groups.values())

    def _preprocess(self, input, log_id):
        if input.fileType is None:
            if utils.is_url(input.file):
                maybe_file_type = utils.infer_file_type(input.file)
                if maybe_file_type is None or not (
                    maybe_file_type == "PDF" or maybe_file_type == "IMAGE"
                ):
                    return protocol.create_aistudio_output_without_result(
                        422,
                        "Unsupported file type",
                        log_id=log_id,
                    )
                file_type = maybe_file_type
            else:
                return protocol.create_aistudio_output_without_result(
                    422,
                    "File type cannot be determined",
                    log_id=log_id,
                )
        else:
            file_type = "PDF" if input.fileType == 0 else "IMAGE"
        visualize_enabled = (
            input.visualize
            if input.visualize is not None
            else self.app_config.visualize
        )

        try:
            file_bytes = utils.get_raw_bytes(input.file)
            images, data_info = utils.file_to_images(
                file_bytes,
                file_type,
                max_num_imgs=self.context["max_num_input_imgs"],
            )
        except Exception as e:
            logging.error("Failed to get input file bytes: %s", e)
            return protocol.create_aistudio_output_without_result(
                422,
                "Input file is invalid",
                log_id=log_id,
            )

        return images, data_info, visualize_enabled

    def _postprocess(self, images, data_info, visualize_enabled, preds, log_id, input):
        ocr_results: List[Dict[str, Any]] = []
        for i, (img, item) in enumerate(zip(images, preds)):
            pruned_res = app_common.prune_result(item.json["res"])
            if visualize_enabled:
                output_imgs = item.img
                imgs = {
                    "input_img": img,
                    "ocr_img": output_imgs["ocr_res_img"],
                }
                if "preprocessed_img" in output_imgs:
                    imgs["doc_preprocessing_img"] = output_imgs["preprocessed_img"]
                imgs = app_common.postprocess_images(
                    imgs,
                    log_id,
                    filename_template=f"{{key}}_{i}.jpg",
                    file_storage=self.context["file_storage"],
                    return_urls=self.context["return_img_urls"],
                    max_img_size=self.context["max_output_img_size"],
                )
            else:
                imgs = {}
            ocr_results.append(
                dict(
                    prunedResult=pruned_res,
                    ocrImage=imgs.get("ocr_img"),
                    docPreprocessingImage=imgs.get("doc_preprocessing_img"),
                    inputImage=imgs.get("input_img"),
                )
            )

        return schemas.ocr.InferResult(
            ocrResults=ocr_results,
            dataInfo=data_info,
        )



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from .contrib.yolov7 import *
from .contrib.yolor import YOLOR
from .contrib.scaled_yolov4 import ScaledYOLOv4
from .contrib.nanodet_plus import NanoDetPlus
from .contrib.yolox import YOLOX
from .contrib.yolov5 import *
from .contrib.yolov5seg import *
from .contrib.fastestdet import *
from .contrib.yolov5lite import YOLOv5Lite
from .contrib.yolov6 import YOLOv6
from .contrib.yolov7end2end_trt import YOLOv7End2EndTRT
from .contrib.yolov7end2end_ort import YOLOv7End2EndORT
from .contrib.yolov8 import *
from .ppdet import *
from .contrib.rkyolo import *



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...common.transformers.transformers import PretrainedConfig

DEFAULT_CONFIG = {
    "model_type": "det",
    "model_name": "PP-OCRv5_server_det",
    "algorithm": "DB",
    "upsample_mode": "nearest",
    "upsample_align_mode": 1,
    "backbone": {
        "name": "PPHGNetV2",
        "stem_channels": [3, 32, 48],
        "stage_config": {
            "stage1": [48, 48, 128, 1, False, False, 3, 6, 2],
            "stage2": [128, 96, 512, 1, True, False, 3, 6, 2],
            "stage3": [512, 192, 1024, 3, True, True, 5, 6, 2],
            "stage4": [1024, 384, 2048, 1, True, True, 5, 6, 2],
        },
        "use_lab": False,
        "use_last_conv": True,
        "class_expand": 2048,
        "dropout_prob": 0.0,
        "class_num": 1000,
        "lr_mult_list": [1.0, 1.0, 1.0, 1.0, 1.0],
        "det": True,
        "out_indices": [0, 1, 2, 3],
    },
    "neck": {
        "name": "LKPAN",
        "out_channels": 256,
        "mode": "large",
        "reduce_factor": 2,
        "intraclblock_config": {
            "reduce_channel": [1, 1, 0],
            "return_channel": [1, 1, 0],
            "v_layer_7x1": [[7, 1], [1, 1], [3, 0]],
            "v_layer_5x1": [[5, 1], [1, 1], [2, 0]],
            "v_layer_3x1": [[3, 1], [1, 1], [1, 0]],
            "q_layer_1x7": [[1, 7], [1, 1], [0, 3]],
            "q_layer_1x5": [[1, 5], [1, 1], [0, 2]],
            "q_layer_1x3": [[1, 3], [1, 1], [0, 1]],
            "c_layer_7x7": [[7, 7], [1, 1], [3, 3]],
            "c_layer_5x5": [[5, 5], [1, 1], [2, 2]],
            "c_layer_3x3": [[3, 3], [1, 1], [1, 1]],
        },
    },
    "head": {
        "name": "PFHeadLocal",
        "in_channels": 1024,
        "k": 50,
        "mode": "large",
        "scale_factor": 2,
        "act": "relu",
        "kernel_list": [3, 2, 2],
        "fix_nan": False,
    },
}


class PPOCRV5ServerDetConfig(PretrainedConfig):
    model_type = "det"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.model_name = kwargs.get("model_name", DEFAULT_CONFIG["model_name"])
        self.algorithm = kwargs.get("algorithm", DEFAULT_CONFIG["algorithm"])
        self.upsample_mode = kwargs.get(
            "upsample_mode", DEFAULT_CONFIG["upsample_mode"]
        )
        self.upsample_align_mode = kwargs.get(
            "upsample_align_mode", DEFAULT_CONFIG["upsample_align_mode"]
        )

        backbone_cfg = kwargs.get("backbone", DEFAULT_CONFIG["backbone"])
        self.backbone_name = backbone_cfg.get(
            "name", DEFAULT_CONFIG["backbone"]["name"]
        )
        self.backbone_stem_channels = backbone_cfg.get(
            "stem_channels", DEFAULT_CONFIG["backbone"]["stem_channels"]
        )
        self.backbone_stage_config = backbone_cfg.get(
            "stage_config", DEFAULT_CONFIG["backbone"]["stage_config"]
        )
        self.backbone_use_lab = backbone_cfg.get(
            "use_lab", DEFAULT_CONFIG["backbone"]["use_lab"]
        )
        self.backbone_use_last_conv = backbone_cfg.get(
            "use_last_conv", DEFAULT_CONFIG["backbone"]["use_last_conv"]
        )
        self.backbone_class_expand = backbone_cfg.get(
            "class_expand", DEFAULT_CONFIG["backbone"]["class_expand"]
        )
        self.backbone_class_num = backbone_cfg.get(
            "class_num", DEFAULT_CONFIG["backbone"]["class_num"]
        )
        self.backbone_lr_mult_list = backbone_cfg.get(
            "lr_mult_list", DEFAULT_CONFIG["backbone"]["lr_mult_list"]
        )
        self.backbone_det = backbone_cfg.get("det", DEFAULT_CONFIG["backbone"]["det"])
        self.backbone_out_indices = backbone_cfg.get(
            "out_indices", DEFAULT_CONFIG["backbone"]["out_indices"]
        )

        neck_cfg = kwargs.get("neck", DEFAULT_CONFIG["neck"])
        self.neck_name = neck_cfg.get("name", DEFAULT_CONFIG["neck"]["name"])
        self.neck_out_channels = neck_cfg.get(
            "out_channels", DEFAULT_CONFIG["neck"]["out_channels"]
        )
        self.neck_mode = neck_cfg.get("mode", DEFAULT_CONFIG["neck"]["mode"])
        self.neck_reduce_factor = neck_cfg.get(
            "reduce_factor", DEFAULT_CONFIG["neck"]["reduce_factor"]
        )
        self.neck_intraclblock_config = neck_cfg.get(
            "intraclblock_config", DEFAULT_CONFIG["neck"]["intraclblock_config"]
        )

        head_cfg = kwargs.get("head", DEFAULT_CONFIG["head"])
        self.head_name = head_cfg.get("name", DEFAULT_CONFIG["head"]["name"])
        self.head_in_channels = head_cfg.get(
            "in_channels", DEFAULT_CONFIG["head"]["in_channels"]
        )
        self.head_k = head_cfg.get("k", DEFAULT_CONFIG["head"]["k"])
        self.head_mode = head_cfg.get("mode", DEFAULT_CONFIG["head"]["mode"])
        self.head_scale_factor = head_cfg.get(
            "scale_factor", DEFAULT_CONFIG["head"]["scale_factor"]
        )
        self.head_act = head_cfg.get("act", DEFAULT_CONFIG["head"]["act"])
        self.head_kernel_list = head_cfg.get(
            "kernel_list", DEFAULT_CONFIG["head"]["kernel_list"]
        )
        self.head_fix_nan = head_cfg.get("fix_nan", DEFAULT_CONFIG["head"]["fix_nan"])



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
from typing import List

from .hub_config import config


class ServerConnectionError(Exception):
    def __init__(self, url: str):
        self.url = url

    def __str__(self):
        tips = "Can't connect to UltraInfer Model Server: {}".format(self.url)
        return tips


class ModelServer(object):
    """
    UltraInfer server source

    Args:
        url(str) : Url of the server
        timeout(int) : Request timeout
    """

    def __init__(self, url: str, timeout: int = 10):
        self._url = url
        self._timeout = timeout

    def search_model(
        self, name: str, format: str = None, version: str = None
    ) -> List[dict]:
        """
        Search model from model server.

        Args:
            name(str) : UltraInfer model name
            format(str): UltraInfer model format
            version(str) : UltraInfer model version
        Return:
            result(list): search results
        """
        params = {}
        params["name"] = name
        if format:
            params["format"] = format
        if version:
            params["version"] = version
        result = self.request(path="ultra_infer_search", params=params)
        if result["status"] == 0 and len(result["data"]) > 0:
            return result["data"]
        return None

    def stat_model(self, name: str, format: str, version: str):
        """
        Note a record when download a model for statistics.

        Args:
            name(str) : UltraInfer model name
            format(str): UltraInfer model format
            version(str) : UltraInfer model version
        Return:
            is_successful(bool): True if successful, False otherwise
        """
        params = {}
        params["name"] = name
        params["format"] = format
        params["version"] = version
        params["from"] = "ultra_infer"
        try:
            result = self.request(path="stat", params=params)
        except Exception:
            return False
        if result["status"] == 0:
            return True
        else:
            return False

    def request(self, path: str, params: dict) -> dict:
        """Request server."""
        api = "{}/{}".format(self._url, path)
        try:
            result = requests.get(api, params, timeout=self._timeout)
            return result.json()
        except requests.exceptions.ConnectionError as e:
            raise ServerConnectionError(self._url)

    def get_model_list(self):
        """
        Get all pre-trained models information in dataset.
        Return:
            result(dict): key is category name, value is a list which contains models \
                information such as name, format and version.
        """
        api = "{}/{}".format(self._url, "ultra_infer_listmodels")
        try:
            result = requests.get(api, timeout=self._timeout)
            return result.json()
        except requests.exceptions.ConnectionError as e:
            raise ServerConnectionError(self._url)

    def is_connected(self):
        return self.check(self._url)

    @classmethod
    def check(cls, url: str) -> bool:
        """
        Check if the specified url is a valid model server

        Args:
            url(str) : Url to check
        """
        try:
            r = requests.get(url + "/search")
            return r.status_code == 200
        except:
            return False


model_server = ModelServer(config.server)



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddlex import create_pipeline

pipeline = create_pipeline(pipeline="object_detection")

output = pipeline.predict(
    "./test_samples/general_layout.png",
    threshold={0: 0.45, 2: 0.48, 7: 0.4},
    layout_nms=True,
    layout_merge_bboxes_mode="large",
    layout_unclip_ratio=(1.0, 1.0),
)

# output = pipeline.predict(
#     "./test_samples/general_layout.png",
# )

# output = pipeline.predict(
#     "./test_samples/general_layout.png",
#     threshold={0: 0.45, 2: 0.48, 7: 0.4},
#     layout_nms=False,
#     layout_merge_bboxes_mode="small",
#     layout_unclip_ratio=1.1
# )

for res in output:
    print(res)
    res.print()  ## 打印预测的结构化输出
    res.save_to_img("./output/")  ## 保存结果可视化图像
    res.save_to_json("./output/")  ## 保存预测的结构化输出
